import unittest
import numpy as np

from .array_comparison import assert_arrays_almost_equal


class ArrayComparisonTest(unittest.TestCase):
	def test_arrays_almost_equal(self):
		a = [1, 2, 3]
		with self.assertRaises(AssertionError):
			assert_arrays_almost_equal(a, [0, 0, 0])

		# Should not raise an error:
		assert_arrays_almost_equal(a, [1, 2, 3])

		# Should not raise an error:
		assert_arrays_almost_equal(a, [1.01, 2.01, 3.01], decimals=1)


if __name__ == '__main__':
	unittest.main()



import numpy as np

from .arithmetic import skew


# Cached for efficiency...
_eye3 = np.eye(3)


def exp(v):
    """
    Compute a rotation of ||v|| radians about the axis v.
    This equals the matrix exponential of the skew-symmetric matrix of v.
    This is also known as the mapping from the Lie algebra so(3) to the Lie group SO(3).
    """
    v = np.asarray(v)
    assert np.shape(v) == (3,), 'exp() received shape %s' % str(v.shape)

    vsq = np.dot(v, v)
    if vsq < 1e-8:
        # Taylor expansion of sin(sqrt(v))/sqrt(v):
        #   http://www.wolframalpha.com/input/?i=sin(sqrt(v))/sqrt(v)
        a = 1. - vsq/6. + vsq*vsq/120.

        # Taylor expansion of (1 - cos(sqrt(v))/v:
        #   http://www.wolframalpha.com/input/?i=(1-cos(sqrt(v)))/v
        b = .5 - vsq/24. + vsq*vsq/720.
    else:
        t = np.sqrt(vsq)
        a = np.sin(t)/t
        b = (1. - np.cos(t)) / vsq

    sk = skew(v)
    return _eye3 + a*sk + b*np.dot(sk, sk)


def exp_jacobian(v):
    """
    Compute the jacobian of exp(v) w.r.t. v. This function return a 3x3 matrix R such that
    for infinitesimal 3-vectors dx, exp(R * dx) * exp(v) = exp(v + dx)
    """
    vsq = np.dot(v, v)
    if vsq < 1e-8:
        # Taylor expansion:
        # http://www.wolframalpha.com/input/?i=2+*+sin%28x%2F2%29+*+sin%28x%2F2%29+%2F+%28x*v%29
        a = .5 - vsq/24 + vsq*vsq/720
        # Taylor expansion:
        # http://www.wolframalpha.com/input/?i=%28x-sin%28x%29%29%2Fx%5E3
        b = 1./6. - vsq/120 + vsq*vsq/5040
    else:
        t = np.sqrt(vsq)
        a = -2. * np.sin(t/2.) * np.sin(t/2.) / vsq
        b = (t - np.sin(t)) / (t*t*t)

    sk = skew(v)
    return np.transpose(_eye3 + a * sk + b * np.dot(sk, sk))


def log(q):
    """
    Compute the axis-angle representation of a rotation.
    If the return value of this function is v then the matrix logarithm of Q is skew(v).
    This is also known as the mapping from the Lie group SO(3) to the Lie algebra so(3).
    """
    q = np.asarray(q)
    assert np.shape(q) == (3, 3), 'log() received shape %s' % str(np.shape(q))

    # http://math.stackexchange.com/questions/83874/
    t = float(q.trace())
    v = np.array((q[2, 1] - q[1, 2],
                  q[0, 2] - q[2, 0],
                  q[1, 0] - q[0, 1]))
    if t >= 3. - 1e-8:
        return (.5 - (t-3.)/12.) * v
    elif t > -1. + 1e-8:
        th = np.arccos(t/2. - .5)
        return th / (2. * np.sin(th)) * v
    else:
        assert t <= -1. + 1e-8, 't=%f, R=%s' % (t, q)
        a = int(np.argmax(q[np.diag_indices_from(q)]))
        b = (a+1) % 3
        c = (a+2) % 3
        s = np.sqrt(q[a,a] - q[b, b] - q[c, c] + 1.)
        v = np.empty(3)
        v[a] = s/2.
        v[b] = (q[b, a] + q[a, b]) / (2.*s)
        v[c] = (q[c, a] + q[a, c]) / (2.*s)
        return v / np.linalg.norm(v)


def perturb_left(q, dq):
    """
    Compute exp(dq)*q where q is a 3x3 rotation and dq is an axis-angle vector.
    """
    return np.dot(exp(dq), q)


def displacement_left(r1, r2):
    """
    Find a vector v such that exp(v) * r1 = r2.
    """
    return log(np.dot(r2, r1.T))


def displacement_left_jacobian_wrt_lhs(r1, r2):
    # TODO: implement analytically
    from manifolds import numeric_jacobian
    return numeric_jacobian(lambda q: displacement_left(q, r2), r1, atlas=RotationAtlas)


def displacement_left_jacobian_wrt_rhs(r1, r2):
    # TODO: implement analytically
    from manifolds import numeric_jacobian
    return numeric_jacobian(lambda q: displacement_left(r1, q), r2, atlas=RotationAtlas)


def angular_velocity_from_axisangle_rates(x, dxdt):
    """
    Compute the derivative of exp(x(t)) w.r.t v, where dxdt is the derivative of x(.) w.r.t. t.
    """
    return -np.dot(exp_jacobian(x).T, dxdt)


class RotationAtlas(object):
    """
    Represents a mapping from 3D rotations to a local parameterization for rotations.
    """
    @classmethod
    def perturb(cls, q, delta):
        return perturb_left(q, delta)

    @classmethod
    def displacement(cls, r1, r2):
        return displacement_left(r1, r2)

    @classmethod
    def dof(cls, _):
        return 3


def unroll_axisangle(a, ref):
    """
    Out of all the axisangle vectors representing the same rotation as A, return the one closest to REF, where
    closest means L2 norm between axisangle vectors.

    Note that if A is a unit vector and t is in radians then the axisangle vector t*a represents the same rotation as
    a*(t + 2*pi), a*(t + 4*pi), etc. So if v is an axisangle vector then v represents the same rotation as
    v*(1. + 2*pi/norm(a)), v*(1. + 4*pi/norm(a)), etc.
    """
    num_full_circles = np.round(np.dot(ref - a, a) / (2. * np.pi * np.linalg.norm(a)))
    return a + num_full_circles * 2 * np.pi * a / np.linalg.norm(a)



from .arithmetic import \
	pr, unpr, normalized, unreduce, unreduce_2d, sumsq, skew, \
	unit, orthonormalize, minmedmax, cis, dots, spy

from .transform import SO3, SE3

from .array_comparison import assert_arrays_almost_equal



import unittest
import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline

from manifolds import numeric_jacobian

from .array_comparison import assert_arrays_almost_equal
from . import rotation


def angular_velocity_global_numerical(r0, r1, delta_t):
    return rotation.log(np.dot(r0.T, r1)) / delta_t


def angular_velocity_local_numerical(r0, r1, delta_t):
    return rotation.log(np.dot(r1, r0.T)) / delta_t


class RotationTest(unittest.TestCase):
    def test_angular_velocity_from_axisangle_rates(self):
        np.random.seed(0)

        # create three splines
        timestamps = np.linspace(0, 10, 10)
        controls = np.random.rand(3, len(timestamps))
        fs = [InterpolatedUnivariateSpline(timestamps, xs) for xs in controls]

        t = 0
        t0 = t - 1e-8
        t1 = t + 1e-8
        r = rotation.exp([f(t) for f in fs])
        r0 = rotation.exp([f(t0) for f in fs])
        r1 = rotation.exp([f(t1) for f in fs])

        w_numerical_global = angular_velocity_global_numerical(r0, r1, t0-t1)
        w_numerical_local = angular_velocity_local_numerical(r0, r1, t0-t1)

        ff = np.array([f(t) for f in fs])
        dfdt = np.array([f.derivative()(t) for f in fs])
        w_analytic_global = rotation.angular_velocity_from_axisangle_rates(ff, dfdt)
        w_analytic_local = np.dot(r, w_analytic_global)

        assert_arrays_almost_equal(w_numerical_global, w_analytic_global, xlabel='numerical', ylabel='analytic')
        assert_arrays_almost_equal(w_numerical_local, w_analytic_local, xlabel='numerical', ylabel='analytic')

    def test_angular_velocity_from_axisangle_rates_at_zero(self):
        np.random.seed(0)

        # create three splines
        timestamps = np.linspace(0, 10, 10)
        controls = np.zeros((3, len(timestamps)))
        fs = [InterpolatedUnivariateSpline(timestamps, xs) for xs in controls]

        t = 0
        t0 = t - 1e-8
        t1 = t + 1e-8
        r = rotation.exp([f(t) for f in fs])
        r0 = rotation.exp([f(t0) for f in fs])
        r1 = rotation.exp([f(t1) for f in fs])

        w_numerical_global = angular_velocity_global_numerical(r0, r1, t0-t1)
        w_numerical_local = angular_velocity_local_numerical(r0, r1, t0-t1)

        ff = np.array([f(t) for f in fs])
        dfdt = np.array([f.derivative()(t) for f in fs])
        w_analytic_global = rotation.angular_velocity_from_axisangle_rates(ff, dfdt)
        w_analytic_local = np.dot(r, w_analytic_global)

        assert_arrays_almost_equal(w_numerical_global, w_analytic_global, xlabel='numerical', ylabel='analytic')
        assert_arrays_almost_equal(w_numerical_local, w_analytic_local, xlabel='numerical', ylabel='analytic')

    def test_exp_jacobian(self):
        np.random.seed(0)
        x0 = np.random.randn(3)
        j_numerical = numeric_jacobian(lambda x: rotation.exp(x), x0,
                                       output_atlas=rotation.RotationAtlas)
        j_analytic = rotation.exp_jacobian(x0)
        np.testing.assert_array_almost_equal(j_numerical, j_analytic)


    def test_displacement_left_jacobian_wrt_lhs(self):
        np.random.seed(0)
        r1 = rotation.exp(np.random.randn(3))
        r2 = rotation.exp(np.random.randn(3))
        j_numerical = numeric_jacobian(lambda r: rotation.displacement_left(r, r2),
                                       r1,
                                       atlas=rotation.RotationAtlas)
        j_analytic = rotation.displacement_left_jacobian_wrt_lhs(r1, r2)
        assert_arrays_almost_equal(j_numerical, j_analytic, xlabel='analytic', ylabel='numerical')


if __name__ == '__main__':
    np.random.seed(123)
    unittest.main()



import numpy as np

from .arithmetic import spy


def assert_arrays_almost_equal(x, y, decimals=6, linelimit=20, xlabel='LHS', ylabel='RHS'):
    """
    Check that x and y are equal to within the specified number of decimal places.
    If they are not then print some helpful diagnostic information and raise an AssertionError.
    """
    x = np.asarray(x)
    y = np.asarray(y)
    assert x.shape == y.shape, 'Shape mismatch: %s vs %s' % ('x'.join(map(str, x.shape)), 'x'.join(map(str, y.shape)))
    tol = np.maximum(10**-decimals, np.abs(np.maximum(x, y) * 10**-decimals))
    mismatches = np.abs(x - y) > tol
    if np.any(mismatches):
        error_coords = list(zip(*np.nonzero(mismatches)))
        for coord in error_coords[:linelimit]:
            print('Mismatch at position %s: %s=%s vs %s=%s' % (','.join(map(str, coord)), xlabel, x[coord], ylabel, y[coord]))
        if len(error_coords) > linelimit:
            print('... and %d more' % (len(error_coords) - linelimit))
        if x.ndim > 1 and np.prod(x.shape) > 10:
            print('\n' + xlabel + ':')
            print(spy(x))
            print('\n' + ylabel + ':')
            print(spy(y))
            print('\nError:')
            print(spy(mismatches))
        else:
            print('\n:' + xlabel + ':')
            print(x)
            print('\n' + ylabel + ':')
            print(y)
            print('\nSparsity pattern of difference:')
            print(spy(mismatches))
        raise AssertionError('arrays were not equal to %d decimals' % decimals)



import unittest
import numpy as np

from .array_comparison import assert_arrays_almost_equal
from .transform import SO3, SE3
from . import rotation


class SO3Test(unittest.TestCase):
    def test_identity(self):
        r = SO3.identity()
        assert_arrays_almost_equal(r.matrix, np.eye(3))

    def test_from_tangent(self):
        r = SO3.from_tangent([.1, .2, .3])
        assert_arrays_almost_equal(r.matrix, rotation.exp([.1, .2, .3]))

    def test_inverse(self):
        r = SO3.from_tangent([.1, .2, .3])
        assert_arrays_almost_equal(r.inverse().matrix, r.matrix.T)

    def test_log(self):
        r = SO3.from_tangent([.1, .2, .3])
        assert_arrays_almost_equal(r.log(), [.1, .2, .3])

class SE3Test(unittest.TestCase):
    def test_identity(self):
        r = SE3.identity()
        assert_arrays_almost_equal(r.matrix[:3, :3], np.eye(3))
        assert_arrays_almost_equal(r.matrix[:3, 3], [0, 0, 0])
        assert_arrays_almost_equal(r.matrix[3], [0, 0, 0, 1])

    def test_from_tangent(self):
        r = SE3.from_tangent([.1, .2, .3, -1, -2, -3])
        assert_arrays_almost_equal(r.orientation, rotation.exp([.1, .2, .3]))
        assert_arrays_almost_equal(r.position, [-1, -2, -3])

    def test_inverse(self):
        r = SE3.from_tangent([.1, .2, .3, -1, -2, -3])
        rinv = SE3(r.orientation.T, -np.dot(r.orientation, r.position))
        assert_arrays_almost_equal(r.inverse().matrix, rinv.matrix)

    def test_log(self):
        r = SE3.from_tangent([.1, .2, .3, -1, -2, -3])
        assert_arrays_almost_equal(r.log(), [.1, .2, .3, -1, -2, -3])

    def test_from_matrix(self):
        r = rotation.exp([.1, .2, .3])
        t = np.array([1., 2., 3.])
        v = SE3.from_matrix(np.hstack((r, t[:, None])))
        assert_arrays_almost_equal(v.orientation, r)
        assert_arrays_almost_equal(v.position, -np.dot(r, t))

    def test_rt(self):
        r = rotation.exp([.1, .2, .3])
        t = np.array([1., 2., 3.])
        v = SE3.from_matrix(np.hstack((r, t[:, None])))
        rr, tt = v.rt
        assert_arrays_almost_equal(r, rr)
        assert_arrays_almost_equal(t, tt)

    def test_rp(self):
        r = rotation.exp([.1, .2, .3])
        p = np.array([1., 2., 3.])
        v = SE3(r, p)
        rr, pp = v.rp
        assert_arrays_almost_equal(r, rr)
        assert_arrays_almost_equal(p, pp)


if __name__ == '__main__':
    unittest.main()


# class SE3(object):
#     """
#     Represents a rigid transform in three dimensions.
#     """
#     DoF = 6

#     class Atlas(object):
#         """
#         Represents an atlas for rigid transforms.
#         """
#         @classmethod
#         def dof(cls, _):
#             return SE3.DoF

#         @classmethod
#         def perturb(cls, pose, tangent):
#             """
#             Evaluate the chart for the given pose at tangent.
#             """
#             assert len(tangent) == SE3.DoF
#             return SE3(rotation.perturb_left(pose.orientation, tangent[:3]), pose.position + tangent[3:])

#         @classmethod
#         def displacement(cls, x1, x2):
#             """
#             Get a vector v such that perturb(x1, v) = x2.
#             """
#             return np.hstack((rotation.log(np.dot(x2.orientation, x1.orientation.T)), x2.position - x1.position))

#     def __init__(self, orientation, position):
#         """
#         Initialize a rigid body transform from a rotation matrix and position vector.
#         """
#         self._orientation = np.asarray(orientation, float)
#         self._position = np.asarray(position, float)

#     @classmethod
#     def identity(cls):
#         """
#         Get the identity transform.
#         """
#         return SE3(np.eye(3), np.zeros(3))

#     @classmethod
#     def from_tangent(cls, v):
#         """
#         Construct a rigid body transform from the tangent space at the identity element.
#         """
#         assert len(v) == SE3.DoF
#         return SE3(rotation.exp(v[:3]), v[3:])

#     @property
#     def orientation(self):
#         """
#         Get the orientation component of this transform.
#         """
#         return self._orientation

#     @orientation.setter
#     def orientation(self, v):
#         """
#         Set the orientation component of this transform.
#         """
#         self._orientation = v

#     @property
#     def position(self):
#         """
#         Get the position component of this transform.
#         """
#         return self._position

#     @position.setter
#     def position(self, v):
#         """
#         Set the position component of this transform.
#         """
#         self._position = v

#     @property
#     def matrix(self):
#         """
#         Get the matrix representation of this transform.
#         """
#         return np.r_[np.c_[self._orientation, -np.dot(self._orientation, self._position)],
#                      np.c_[0., 0., 0., 1.]]

#     @property
#     def rp(self):
#         """
#         Get the (rotation, position) pair for this transform.
#         """
#         return self._orientation, self._position

#     def __mul__(self, rhs):
#         """
#         Multiply this transform with another.
#         """
#         return self.transform(rhs)

#     def transform(self, rhs):
#         """
#         Multiply this transform with another.
#         """
#         if isinstance(rhs, SE3):
#             r1, r2 = self._orientation, rhs._orientation
#             return SE3(np.dot(r1, r2), rhs.position + np.dot(r2.T, self.position))
#         elif isinstance(rhs, np.ndarray):
#             if rhs.shape[-1] == 3:
#                 return np.dot(self._orientation, rhs - self.position)
#             elif rhs.shape[-1] == 4:
#                 return np.dot(self.matrix, rhs)

#     def inverse(self):
#         """
#         Get the inverse of this transform.
#         """
#         return SE3(self._orientation.T, -np.dot(self.orientation, self.position))

#     def __str__(self):
#         """
#         Get a string representation of this transform.
#         """
#         return 'SE3(position=%s, log_rotation=%s)' % (self._position, rotation.log(self._orientation))



import unittest
import numpy as np
from numpy.testing import assert_array_almost_equal

from . import arithmetic

class ArithmeticTest(unittest.TestCase):
	def test_normalized(self):
		a = [2., 0., 0.]
		assert_array_almost_equal(arithmetic.normalized(a), [1., 0., 0.])

	def test_pr(self):
		a = [4., 6., 2.]
		assert_array_almost_equal(arithmetic.pr(a), [2., 3.])

	def test_unpr(self):
		a = [1., 2.]
		assert_array_almost_equal(arithmetic.unpr(a), [1., 2., 1.])

	def test_spy(self):
		a = [[1, 0], [1, 1]]
		self.assertEqual(arithmetic.spy(a), '[x ]\n[xx]')

	def test_unreduce(self):
		a = [1, 2]
		mask = [False, True, False, True, False]
		assert_array_almost_equal(arithmetic.unreduce(a, mask), [0, 1, 0, 2, 0])

	def test_unreduce_2d(self):
		a = [[1, 2], [3, 4]]
		mask = [True, False, True]
		expected = [
			[1, 0, 2],
			[0, 0, 0],
			[3, 0, 4]]
		assert_array_almost_equal(arithmetic.unreduce_2d(a, mask), expected)

	def test_dots(self):
		a = np.eye(2)
		b = np.array([[1., 2.], [3., 4.]])
		assert_array_almost_equal(arithmetic.dots(a, b, a), b)

	def test_sumsq(self):
		a = [1, 2]
		assert_array_almost_equal(arithmetic.sumsq(a), 5)

	def test_skew(self):
		a = [1, 2, 3]
		expected = [
			[0, -3, 2],
			[3, 0, -1],
			[-2, 1, 0]]
		assert_array_almost_equal(arithmetic.skew(a), expected)

	def test_unit(self):
		assert_array_almost_equal(arithmetic.unit(1, 5), [0, 1, 0, 0, 0])

	def test_min_med_max(self):
		a = [1, 2, 3, 4, 5]
		assert_array_almost_equal(arithmetic.minmedmax(a), [1, 3, 5])




if __name__ == '__main__':
	unittest.main()



import functools
import numpy as np


def pr(a):
    """
    For vectors A, divide each element of A by the last.
    For matrices, divide each row by its rightmost element.
    For arrays of high dimension, project along the last dimension.
    """
    a = np.asarray(a)
    return a[..., :-1] / a[..., -1:]


def unpr(a):
    """
    For vectors A, return A with 1 appended.
    For matrices, append a column of ones.
    For arrays of higher dimension, append along the last dimension.
    """
    a = np.asarray(a)
    col_shape = a.shape[:-1] + (1,)
    return np.concatenate((a, np.ones(col_shape)), axis=-1)


def normalized(a):
    """
    For vertors A, return a unit vector in the direction of A.
    For matrices, return a matrix of the same size where each row has unit norm.
    For arrays of higher dimension, normalize along the last axis.
    """
    a = np.asarray(a)
    return a / np.sqrt(np.sum(np.square(a), axis=-1))[..., None]


def unreduce(a, mask, fill=0.):
    """
    Create a vector in which the positions set to True in the mask vector are set to sequential values from A.
    """
    a = np.asarray(a)
    mask = np.asarray(mask)
    x = np.repeat(fill, len(mask)).astype(a.dtype)
    x[mask] = a
    return x


def unreduce_2d(a, mask, fill=0.):
    """
    Create a matrix in which the rows and columns set to True in the mask vector are set to the entries from A.
    """
    a = np.asarray(a)
    x = np.ones((len(mask), len(mask))) * fill
    x[np.ix_(mask, mask)] = a
    return x


def sumsq(a, axis=None):
    """
    Compute the sum of squares.
    """
    return np.sum(np.square(a), axis=axis)


def skew(a):
    """
    Compute the skew-symmetric matrix for a. The returned matrix M has the property 
    that for any 3-vector v, M * v = a x v where x denotes the cross product.
    """
    a = np.asarray(a)
    assert a.shape == (3,), 'shape was was %s' % str(a.shape)
    return np.array([[0, -a[2], a[1]],
                     [a[2], 0, -a[0]],
                     [-a[1], a[0], 0.]])


def unit(i, n):
    """
    Compute a unit vector along the i-th axis in N dimensions.
    """
    return (np.arange(n) == i).astype(float)


def orthonormalize(a):
    """
    Find the closest orthonormal matrix to R.
    Note that the returned matrix may have determinant +1 or -1, so the result
    may be either a rotation or a rotoinversion.
    """
    u, _, v = np.linalg.svd(a)
    return np.dot(u, v)


def minmedmax(a):
    """
    Compute [min(x), median(a), max(a)]
    """
    if len(a) == 0:
        raise Exception('warning [utils.minmedmax]: empty list passed')
    else:
        return np.min(a), np.median(a), np.max(a)


def cis(th):
    """
    Compute [cos(th), sin(th)].
    """
    return np.array((np.cos(th), np.sin(th)))


def dots(*a):
    """
    Multiply an arbitrary number of matrices with np.dot.
    """
    return functools.reduce(np.dot, a)


def spy(a, t=1e-5):
    """
    Return a string representation of the sparsity pattern of A.
    """
    a = np.atleast_2d(a)
    return '\n'.join(map(lambda row: '['+''.join('x' if abs(val)>t else ' ' for val in row)+']', a))



#!/usr/bin/env python
from distutils.core import setup

setup(name='rigidbody',
      description='Rigid body transforms in 3D',
      version='0.11',
      author='Alex Flint',
      author_email='alex.flint@gmail.com',
      url='https://github.com/alexflint/rigidbody',
      packages=['rigidbody'],
      package_dir={'rigidbody': '.'},
      )



import numpy as np

from . import rotation


class SO3(object):
    """
    Represents a rotation in three dimensions.
    """
    DoF = 3

    class Atlas(object):
        """
        Represents an atlas for rotations.
        """
        @classmethod
        def dof(cls, _):
            return SO3.DoF

        @classmethod
        def perturb(cls, r, tangent):
            """
            Evaluate the chart for the given rotation.
            """
            assert len(tangent) == SO3.DoF
            return SO3(rotation.perturb_left(r.matrix, tangent))

        @classmethod
        def displacement(cls, r1, r2):
            """
            Get a vector v such that perturb(x1, v) = x2.
            """
            return rotation.log(np.dot(SO3.asarray(r2), SO3.asarray(r1).T))

    def __init__(self, r):
        """
        Initialize from a rotation matrix.
        """
        self._r = r

    @classmethod
    def identity(cls):
        """
        Construct the identity rotation.
        """
        return SO3(np.eye(3))

    @classmethod
    def from_tangent(cls, v):
        """
        Construct a rotation for the tangent space at the identity element.
        """
        assert len(v) == SO3.DoF
        return SO3(rotation.exp(v))

    @classmethod
    def asarray(cls, x):
        """
        Convert an SO3 to a matrix.
        """
        if isinstance(x, SO3):
            return x.matrix
        else:
            return np.asarray(x)

    @property
    def matrix(self):
        """
        Get the matrix representation of this rotation.
        """
        return self._r

    def __mul__(self, rhs):
        """
        Multiply this rotation with another.
        """
        return self.transform(rhs)

    def transform(self, rhs):
        """
        Multiply this rotation with another.
        """
        if isinstance(rhs, SO3):
            return SO3(np.dot(self._r, rhs._r))
        elif isinstance(rhs, np.ndarray):
            return SO3(np.dot(self._r, rhs))

    def inverse(self):
        """
        Get the inverse of this rotation
        """
        return SO3(self._r.T)

    def log(self):
        """
        Compute the axis angle representation of this rotation.
        """
        return rotation.log(self._r)

    def __str__(self):
        """
        Get a string representation of this rotation.
        """
        return 'SO3(%s)' % str(self._r).replace('\n', '\n    ')


class SE3(object):
    """
    Represents a rigid transform in three dimensions.
    """
    DoF = 6

    class Atlas(object):
        """
        Represents an atlas for rigid transforms.
        """
        @classmethod
        def dof(cls, _):
            return SE3.DoF

        @classmethod
        def perturb(cls, pose, tangent):
            """
            Evaluate the chart for the given pose at tangent.
            """
            assert len(tangent) == SE3.DoF
            return SE3(rotation.perturb_left(pose.orientation, tangent[:3]), pose.position + tangent[3:])

        @classmethod
        def displacement(cls, x1, x2):
            """
            Get a vector v such that perturb(x1, v) = x2.
            """
            return np.hstack((rotation.log(np.dot(x2.orientation, x1.orientation.T)), x2.position - x1.position))

    def __init__(self, orientation, position):
        """
        Initialize a rigid body transform from a rotation matrix and position vector.
        """
        self._orientation = np.asarray(orientation, float)
        self._position = np.asarray(position, float)

    @classmethod
    def identity(cls):
        """
        Get the identity transform.
        """
        return SE3(np.eye(3), np.zeros(3))

    @classmethod
    def from_tangent(cls, v):
        """
        Construct a rigid body transform from the tangent space at the identity element.
        """
        assert len(v) == SE3.DoF
        return SE3(rotation.exp(v[:3]), v[3:])

    @classmethod
    def from_matrix(cls, m):
        """
        Construct a rigid body transform from a 3x4 or 4x4 matrix
        """
        m = np.asarray(m)
        assert m.shape in ((3, 4), (4, 4)), 'shape was %s' % str(m.shape)
        r = m[:3, :3]
        t = m[:3, 3]
        return SE3(r, -np.dot(r.T, t))

    @property
    def orientation(self):
        """
        Get the orientation component of this transform.
        """
        return self._orientation

    @orientation.setter
    def orientation(self, v):
        """
        Set the orientation component of this transform.
        """
        self._orientation = v

    @property
    def position(self):
        """
        Get the position component of this transform.
        """
        return self._position

    @position.setter
    def position(self, v):
        """
        Set the position component of this transform.
        """
        self._position = v

    @property
    def matrix(self):
        """
        Get the matrix representation of this transform.
        """
        return np.r_[np.c_[self._orientation, -np.dot(self._orientation, self._position)],
                     np.c_[0., 0., 0., 1.]]

    @property
    def rp(self):
        """
        Get the (rotation, position) pair for this transform.
        """
        return self._orientation, self._position

    @property
    def rt(self):
        """
        Get the (rotation, translation) pair for this transform.
        """
        return self._orientation, -np.dot(self._orientation, self._position)

    def __mul__(self, rhs):
        """
        Multiply this transform with another.
        """
        return self.transform(rhs)

    def transform(self, rhs):
        """
        Multiply this transform with another.
        """
        if isinstance(rhs, SE3):
            r1, r2 = self._orientation, rhs._orientation
            return SE3(np.dot(r1, r2), rhs.position + np.dot(r2.T, self.position))
        elif isinstance(rhs, np.ndarray):
            if rhs.shape[-1] == 3:
                return np.dot(self._orientation, rhs - self.position)
            elif rhs.shape[-1] == 4:
                return np.dot(self.matrix, rhs)

    def inverse(self):
        """
        Get the inverse of this transform.
        """
        return SE3(self._orientation.T, -np.dot(self.orientation, self.position))

    def log(self):
        """
        Map this transform to the Lie algebra se3.
        """
        return np.concatenate((rotation.log(self.orientation), self.position))

    def __str__(self):
        """
        Get a string representation of this transform.
        """
        return 'SE3(position=%s, log_rotation=%s)' % (self._position, rotation.log(self._orientation))
